@@ -1,10 +1,7 @@
+from datetime import datetime
 from pathlib import Path
-from data_engineering_extract_metadata.table_list import get_table_list
-from data_engineering_extract_metadata.utils import (
-    write_json,
-    read_json,
-    prepare_folder,
-)
+
+from data_engineering_extract_metadata.utils import prepare_folder, write_json
 
 
 def create_all_metadata(
@@ -13,102 +10,19 @@ def create_all_metadata(
     title,
     description,
     schema,
-    source_bucket,
-    source_folder,
-    include_op_column=True,
-    include_timestamp_column=True,
-    include_scn_column=True,
-    include_derived_columns=True,
-    include_objects=False,
-    filter_function=None,
-):
-    """Perform the whole process: get table list, filter it and create metadata folder
-
-    Just runs get_table_list and create_metadata_folder - no extra processing
-
-    Parameters
-    ----------
-    connection (cx_Oracle connection):
-        A premade connection to a database.
-
-    save_folder (str or Path):
-        Location for the table list and metadata subfolder.
-
-    title (str):
-        Title for the table list file and metadata subfolder.
-
-    description (str):
-        How to describe the database in the Glue catalogue.
-
-    schema (str):
-        Name of the database schema to get tables from.
-
-    source_bucket (str):
-        The name of the S3 bucket the source data will be stored in.
-
-    source_folder (str):
-        Folder/prefix within the S3 bucket for the relevant source data.
-
-    filter_function (function or None):
-        If you want to filter the list of tables after extracting it, pass a function
-        here. The function should take a single argument: a list of tuples in the form
-        (table name, tablespace). It should return a list of table names (not tuples).
-
-    include_op_column (boolean):
-        If True, adds a column called 'Op' for operation type - I, U or D
-        This data is automatically added by AWS DMS ongoing replication tasks.
-
-    include_timestamp_column (string):
-        If True, adds a column called 'extraction_timestamp'.
-        This data is added by DMS if the source extra connection attribute is set.
-
-    include_scn_column (boolean):
-        If True, adds a column called 'scn'.
-        This data is added by DMS after being extracted from the source Oracle table.
-
-    include_derived_columns (boolean):
-        If True, adds three columns: mojap_current_record, mojap_start_datetime,
-        mojap_end_datetime.
-
-    include_objects (boolean):
-        If True, will include metadata for DB_TYPE_OBJECT columns as array<character>.
-        Leave False to ignore these columns - AWS DMS doesn't extract them.
-    """
-    get_table_list(connection, save_folder, title, schema, filter_function)
-    create_metadata_folder(
-        connection,
-        save_folder,
-        title,
-        description,
-        schema,
-        source_bucket,
-        source_folder,
-        include_op_column,
-        include_timestamp_column,
-        include_scn_column,
-        include_derived_columns,
-        include_objects,
-    )
-
-
-def create_metadata_folder(
-    connection,
-    save_folder,
-    title,
-    description,
-    schema,
+    objects,
     source_bucket,
     source_folder,
     include_op_column=True,
     include_timestamp_column=True,
     include_scn_column=True,
     include_derived_columns=False,
-    include_objects=False,
+    include_object_types=False,
 ):
     """
     Creates a folder of metadata readable by etl_manager's read_database_folder.
 
-    You must already have prepared a list of tables using get_table_list.
+    You must already have prepared a list of object using get_object_list.
 
     Parameters
     ----------
@@ -116,17 +30,20 @@ def create_metadata_folder(
         A premade connection to a database.
 
     save_folder (str or Path):
-        Location of the table_list. Metadata will be saved into a subfolder of this.
+        Location of the object_list. Metadata will be saved into a subfolder of this.
 
     title (str):
         Title for this collection of metadata - files will be saved into a subfolder
-        with this name. This should also be the name of the table list file.
+        with this name. This should also be the name of the object list file.
 
     description (str):
         How to describe the database in the Glue catalogue.
 
     schema (str):
-        Name of the database schema to get tables from.
+        Name of the database schema to get objects from.
+
+    objects (list):
+        A list of tables and materialised views for which to extract metadata.
 
     source_bucket (str):
         The name of the S3 bucket the source data will be stored in.
@@ -144,19 +61,30 @@ def create_metadata_folder(
 
     include_scn_column (boolean):
         If True, adds a column called 'scn'.
-        This data is added by DMS after being extracted from the source Oracle table.
+        This data is added by DMS after being extracted from the source Oracle object.
 
     include_derived_columns (boolean):
         If True, adds three columns: mojap_current_record, mojap_start_datetime,
         mojap_end_datetime.
 
-    include_objects (boolean):
+    include_object_types (boolean):
         If True, will include metadata for DB_TYPE_OBJECT columns as array<character>.
         Leave False to ignore these columns - AWS DMS doesn't extract them.
     """
     # Create destination folder if it doesn't already exist
     destination = Path(save_folder) / title
 
+    objects = [o.upper() for o in objects]
+
+    save_objects(
+        title=title,
+        folder=save_folder,
+        objects=objects,
+        blobs=find_blob_columns(
+            objects=objects, database=schema, connection=connection
+        ),
+    )
+
     # Create database.json file
     create_json_for_database(
         description=description,
@@ -166,30 +94,22 @@ def create_metadata_folder(
         location=destination,
     )
 
-    # Get pre-saved table list
-    tables = read_json(title, save_folder)
-    table_list = tables["tables"]
-
-    # Connect to database and get the table metadata
-    create_json_for_tables(
-        tables=table_list,
+    # Connect to database and get the object metadata
+    create_json_for_objects(
+        objects=objects,
         schema=schema,
         location=destination,
         include_op_column=include_op_column,
         include_timestamp_column=include_timestamp_column,
         include_scn_column=include_scn_column,
         include_derived_columns=include_derived_columns,
-        include_objects=include_objects,
+        include_object_types=include_object_types,
         connection=connection,
     )
 
 
 def create_json_for_database(
-    description: str,
-    name: str,
-    bucket: str,
-    base_folder: str,
-    location="./meta_data",
+    description: str, name: str, bucket: str, base_folder: str, location="./meta_data",
 ):
     """Creates a database.json file suitable for reading with read_database_folder()
 
@@ -207,7 +127,7 @@ def create_json_for_database(
     base_folder (str):
         The folder/prefix within the S3 bucket for the data you want
         Use capitals if the S3 location includes capitals
-        Don't go down to the table level - it should be the level above
+        Don't go down to the object level - it should be the level above
         For example "hmpps/delius/DELIUS_ANALYTICS_PLATFORM".
 
     location (str):
@@ -228,25 +148,25 @@ def create_json_for_database(
     write_json(db, Path(location) / "database.json")
 
 
-def create_json_for_tables(
-    tables: list,
+def create_json_for_objects(
+    objects: list,
     schema: str,
     location="./meta_data",
     include_op_column=True,
     include_timestamp_column=True,
     include_scn_column=True,
     include_derived_columns=False,
-    include_objects=False,
+    include_object_types=False,
     connection=None,
 ):
     """
-    Creates a json file of metadata for each table that's named in tables and has rows
+    Creates a json file of metadata for each object that's named in objects and has rows
     These json files are suitable for reading with read_database_folder().
 
     Parameters
     ----------
-    tables (list):
-        A list of table names from the source database.
+    objects (list):
+        A list of object names from the source database.
 
     schema (str):
         The name of the Oracle schema to get metadata from.
@@ -264,13 +184,13 @@ def create_json_for_tables(
 
     include_scn_column (boolean):
         If True, adds a column called 'scn'.
-        This data is added by DMS after being extracted from the source Oracle table.
+        This data is added by DMS after being extracted from the source Oracle object.
 
     include_derived_columns (boolean):
         If True, adds three columns: mojap_current_record, mojap_start_datetime,
         mojap_end_datetime.
 
-    include_objects (boolean):
+    include_object_types (boolean):
         If True, will include metadata for DB_TYPE_OBJECT columns as array<character>.
         Leave False to ignore these columns - AWS DMS doesn't extract them.
 
@@ -280,7 +200,7 @@ def create_json_for_tables(
     Returns
     -------
     None
-        But creates a json file for each table in the database.
+        But creates a json file for each object in the database.
     """
     cursor = connection.cursor()
 
@@ -288,33 +208,36 @@ def create_json_for_tables(
     not_nullable = find_not_nullable(cursor, schema)
 
     problems = {}
-    for table in tables:
+    for o in objects:
         try:
-            # Get a row to see the columns and check the table has data
+            # Get a row to see the columns and check the object has data
             # 'WHERE ROWNUM <= 1' is Oracle for 'LIMIT 1'
             # fetchone() to see the first row of the query executed
-            cursor.execute(f"SELECT * FROM {schema}.{table} WHERE ROWNUM <= 1")
+            cursor.execute(f"SELECT * FROM {schema}.{o} WHERE ROWNUM <= 1")
             cursor.fetchone()
-            # For Delius purposes we want all tables including those without rows
+            # For Delius purposes we want all objects including those without rows
             # This might not be the case for all metadata
-            metadata = get_table_meta(
+            if is_materialised_view(object_=o):
+                o = rename_materialised_view(name=o)
+
+            metadata = get_object_meta(
                 cursor,
-                table,
+                o,
                 schema,
-                not_nullable.get(table.lower(), []),
+                not_nullable.get(o.lower(), []),
                 include_op_column,
                 include_timestamp_column,
                 include_scn_column,
                 include_derived_columns,
-                include_objects,
+                include_object_types,
             )
-            write_json(metadata, Path(location) / f"{table.lower()}.json")
+            write_json(metadata, Path(location) / f"{o.lower()}.json")
 
         except Exception as e:
-            # Likely errors are that the table has been deleted, marked
+            # Likely errors are that the object has been deleted, marked
             # for deletion, or relies on an external reference you can't access
-            print(f"Problem reading {table} in {schema}")
-            problems[table] = e.args[0].message  # this attribute may be Oracle-only
+            print(f"Problem reading {o} in {schema}")
+            problems[o] = e.args[0].message  # this attribute may be Oracle-only
             continue
 
     # Print the error messages at the end
@@ -322,37 +245,48 @@ def create_json_for_tables(
         print()
         print("ERRORS RAISED")
         for p, e in problems.items():
-            print(f"Error in table {p}: {e}")
+            print(f"Error in object {p}: {e}")
 
     cursor.close()
 
 
-def get_table_meta(
+def is_materialised_view(object_: str) -> bool:
+    if object_[-3:] == "_MV":
+        return True
+    else:
+        return False
+
+
+def rename_materialised_view(name: str) -> str:
+    return name[:-3]
+
+
+def get_object_meta(
     cursor,
-    table: str,
+    object_: str,
     schema: str,
     not_nullable: list,
     include_op_column: bool = True,
     include_timestamp_column: bool = True,
     include_scn_column: bool = True,
     include_derived_columns: bool = False,
-    include_objects: bool = False,
+    include_object_types: bool = False,
 ) -> list:
     """
-    Lists a table's columns, plus any primary key fields and partitions
+    Lists an object's columns, plus any primary key fields and partitions
 
     Parameters
     ----------
     cursor:
         A cursor where .execute has already been run - usually querying
-        "SELECT * FROM {database}.{table} WHERE ROWNUM <= 1".
+        "SELECT * FROM {database}.{object_} WHERE ROWNUM <= 1".
         This will give the cursor a .description attribute with column info.
 
-    table (str):
-        Name of the table.
+    object_ (str):
+        Name of the object.
 
     not_nullable (list):
-        List of column names in this table that aren't nullable.
+        List of column names in this object that aren't nullable.
 
     include_op_column (boolean):
         If True, adds a column called 'Op' for operation type - I, U or D
@@ -364,20 +298,20 @@ def get_table_meta(
 
     include_scn_column (boolean):
         If True, adds a column called 'scn'
-        This data is added by DMS after being extracted from the source Oracle table.
+        This data is added by DMS after being extracted from the source Oracle object.
 
     include_derived_columns (boolean):
         If True, adds three columns: mojap_current_record, mojap_start_datetime,
         mojap_end_datetime.
 
-    include_objects (boolean):
+    include_object_types(boolean):
         If True, will include metadata for DB_TYPE_OBJECT columns as array<character>.
         Leave False to ignore these columns - AWS DMS doesn't extract them.
 
     Returns
     -------
     List of dicts
-        Contains data for all the columns in the table, ready to write to json.
+        Contains data for all the columns in the object, ready to write to json.
     """
     # This lookup is specific to Oracle data types
     # Likely to need separate dictionaries for other data sources
@@ -432,7 +366,7 @@ def get_table_meta(
     # Plus blob columns, which we're intentionally excluding
     skip = ["REF", "ANYDATA", "DB_TYPE_ROWID", "DB_TYPE_BFILE", "DB_TYPE_BLOB"]
 
-    if not include_objects:
+    if not include_object_types:
         skip.append("DB_TYPE_OBJECT")
 
     # Main column info - cursor.description has 7 set columns:
@@ -481,26 +415,26 @@ def get_table_meta(
     if include_derived_columns:
         columns += derived_columns
 
-    primary_keys = get_primary_keys(table=table, cursor=cursor)
-    partition_keys = get_partition_keys(table=table, schema=schema, cursor=cursor)
+    primary_keys = get_primary_keys(object_=object_, cursor=cursor)
+    partition_keys = get_partition_keys(object_=object_, schema=schema, cursor=cursor)
 
     metadata = {
         "$schema": (
             "https://moj-analytical-services.github.io/metadata_schema/table/"
             "v1.4.0.json"
         ),
-        "name": table.lower(),
+        "name": object_.lower(),
         "description": "",
         "data_format": "parquet",
         "columns": columns,
-        "location": f"{table}/",
+        "location": f"{object_}/",
         "partitions": partition_keys,
         "primary_key": primary_keys,
     }
     return metadata
 
 
-def get_primary_keys(table, cursor):
+def get_primary_keys(object_, cursor):
     """Looks through constraints for primary keys, and checks they match colums
     Run as part of get_curated_metadata
     """
@@ -511,28 +445,28 @@ def get_primary_keys(table, cursor):
         "AND cons.constraint_name = cols.constraint_name "
         "AND cons.owner = cols.owner "
         "AND cons.status = 'ENABLED' "
-        "AND cons.table_name = :table_name "
+        "AND cons.table_name = :object_name "
         "ORDER BY cols.table_name, cols.position"
     )
-    cursor.execute(statement, table_name=table)
+    cursor.execute(statement, object_name=object_)
     result = cursor.fetchall()
     if result == []:
-        print(f"No primary key fields found for table {table.lower()}")
+        print(f"No primary key fields found for object {object_.lower()}")
         primary_keys = None
     else:
         primary_keys = [item[0].lower() for item in result]
     return primary_keys
 
 
-def get_partition_keys(schema: str, table: str, cursor) -> list:
-    """Returns a list the partitioning key columns for the given table.
+def get_partition_keys(schema: str, object_: str, cursor) -> list:
+    """Returns a list the partitioning key columns for the given object.
 
     Parameters
     ----------
     schema : str
         Also referred to as "owner" in Oracle databases.
-    table : str
-        The table to get partitioning keys for.
+    object_ : str
+        The object to get partitioning keys for.
     cursor :
         A database Cursor object ready to execute queries with.
     """
@@ -540,12 +474,12 @@ def get_partition_keys(schema: str, table: str, cursor) -> list:
         "SELECT * "
         "FROM SYS.all_part_key_columns "
         "WHERE OWNER = :schema "
-        "AND name = :table_name"
+        "AND name = :object_name"
     )
-    cursor.execute(statement, [schema, table])
+    cursor.execute(statement, [schema, object_])
     result = cursor.fetchall()
     if not result:
-        print(f"No partition keys found for table {table.lower()}")
+        print(f"No partition keys found for object {object_.lower()}")
         return None
     else:
         partition_keys = [item[3].lower() for item in result]
@@ -553,9 +487,9 @@ def get_partition_keys(schema: str, table: str, cursor) -> list:
 
 
 def find_not_nullable(cursor, schema):
-    """Returns a dictionary of tables, each with a list of their non-nullable columns.
+    """Returns a dictionary of objects, each with a list of their non-nullable columns.
 
-    Can't use table descriptions for this as the null_ok field is inaccurate in Delius.
+    Can't use object descriptions for this as the null_ok field is inaccurate in Delius.
 
     Parameters
     ----------
@@ -588,3 +522,64 @@ def find_not_nullable(cursor, schema):
         if con[2] == "P" or con[3].endswith("IS NOT NULL"):
             not_nullable.setdefault(con[0].lower(), []).append(con[1].lower())
     return not_nullable
+
+
+def find_blob_columns(objects: list, database: str, connection) -> list:
+    """Return a list of blob columns that in a specified schema and list of objects.
+
+    Parameters
+    ----------
+    objects (list):
+        List of objects to search for blobs.
+
+    database (str):
+        The name of the database/schema - for example DELIUS_ANALYTICS_PLATFORM.
+
+    connection (database connection object):
+        The database connection to query - for example a cx_Oracle.connect() object.
+    """
+    # Format the table list to put it in the SQL query
+    objects_string = ", ".join([f"'{t}'" for t in objects])
+
+    cursor = connection.cursor()
+    cursor.execute(
+        f"SELECT table_name, column_name "
+        f"FROM all_tab_columns "
+        f"WHERE owner = '{database}' "
+        f"AND table_name IN ({objects_string}) "
+        f"AND data_type = 'BLOB' "
+        f"ORDER BY table_name, column_name"
+    )
+    results = cursor.fetchall()
+    cursor.close()
+
+    blobs = [{"object_name": result[0], "column_name": result[1]} for result in results]
+
+    return blobs
+
+
+def save_objects(title: str, folder: str, objects: list, blobs: list = []):
+    """Saves the objects to json with a timestamp.
+
+    Parameters
+    ----------
+
+    title (str):
+        Name for the object list - used in the json and for the filename
+
+    folder (str):
+        Subfolder to save the json to
+
+    objects (list[str]):
+        List of object names
+    """
+    prepare_folder(folder, clear_content=False)
+
+    output = {
+        "objects_from": title,
+        "extraction_date": datetime.now().isoformat(),
+        "objects": sorted(objects),
+        "blobs": blobs,
+    }
+    write_json(output, Path(folder) / f"{title.lower()}.json")
+    print(f"Saved list of {len(output['objects'])} objects")
